import numpy as np
import matplotlib.pyplot as plt
import warnings
from .read_cluster_lifetime import read_cluster_lifetime
from ..file_managment.save_vars_to_file import save_vars_to_file

def complex_lifetime(FileName: str, FileNum: int, InitialTime: float, FinalTime: float,
                     SpeciesName: str, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
    """This line plot indicates the lifetime for different sizes of complexes. The x-axis is the size of complexes, and the y-axis is the corresponding lifetime in unit of second. If multiple input files are given, the output plot will be the average value of all files and an error bar will also be included.

    Args:
        FileName (str): The path to the '.dat' file, which is usually named as 'transition_matrix_time.dat', representing the histogram data to be analyzed.
        FileNum (int): The number of the total input file. If multiple files are provided, their names should obey the naming rule listed below.
        InitialTime (float): The initial time that users desire to examine. The acceptable range should not be smaller than the starting time or exceed the ending time of simulation.
        FinalTime (float): The final time that users desire to examine. The acceptable range should not be smaller than the value of InitialTime or exceed the ending time of simulation.
        SpeciesName (str): The name of species that users want to examine, which should also be identical with the name written in the input (.inp and .mol) files.
        ShowFig (bool, optional): If True, the plot will be shown; if False, the plot will not be shown. No matter the plot is shown or not, the returns will remain the same. Defaults to True.
        SaveFig (bool, optional): If True, the plot will be saved as a '.png' file in the current directory; if False, the figure will not be saved. Defaults to False.
        SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

    Returns:
        A tuple containing the following:
        - np.ndarray: an array of the size of complexes
        - np.ndarray: an array of the dissociate probabilities when complexes dissociate into smaller complexes
        - np.ndarray: an array of the dissociate probabilities when complexes dissociate into larger complexes

    Notes:
        - If multiple input files are given, the output plot will be the average value of all files and an error bar will also be included.
        - For the dissociate reaction, only the complexes of smaller size dissociating from the original one is counted as dissociate event asymmetrically.
        - For example, if a heptamer dissociates into a tetramer and a trimer, then this event is counted only once, which is heptamer dissociates to trimer.
        - If a single file is provided, the input file should be named as its original name ('transition_matrix_time.dat').
        - If multiple files are provided, the name of the input file should also include serial number as 'transition_matrix_time_X.dat' where X = 1,2,3,4,5...
    """
    warnings.filterwarnings('ignore')
    file_name_head = FileName.split('.')[0]
    file_name_tail = FileName.split('.')[1]
    mean_lifetime = []
    
    #reads through each inputted file
    for matrix_file_number in range(1, FileNum+1):
       
        #determining file name (if there are multiple or none)
        if FileNum == 1: temp_file_name = FileName
        else: temp_file_name = file_name_head + '_' + str(matrix_file_number) + '.' + file_name_tail
        
        #reads cluster lifetime
        ti_lifetime, tf_lifetime, size_list = read_cluster_lifetime(
            temp_file_name, SpeciesName, InitialTime, FinalTime)
        
        #gets mean lifetime for each size
        mean_temp = []
        for index,lifetime_f in enumerate(tf_lifetime):
            lifetime_f = np.delete(
                lifetime_f, range(0, len(ti_lifetime[index])), axis=0) #what does this do???
            mean_temp.append(lifetime_f.mean())
        mean_lifetime.append(mean_temp)
    
    #is it... TRANSPOSING TIME???? yes, yes it is
    mean_lifetime_rev = np.array(mean_lifetime).transpose()
    
    #find means and std devs
    mean = []
    std = []
    if FileNum != 1:
        for index,size in enumerate(mean_lifetime_rev):
            if np.any(size != mean_lifetime_rev[index-1]):
                mean.append(np.nanmean(size))
                std.append(np.nanstd(size))
            else:
                mean.append(mean_lifetime_rev[index-1])
                std.append(mean_lifetime_rev[index-1])     
    else:
        for size in mean_lifetime_rev:
            mean.append(size[0])
                 
    #output variables
    if SaveVars:
        save_vars_to_file({"cmplx_size":size_list,"mean_complex_lifetime":mean,"std":std})
    
    #show figure!!!
    if ShowFig:
        errorbar_color = '#c9e3f6'
        plt.plot(size_list, mean, color='C0')
        if FileNum != 1:
            plt.errorbar(size_list, mean, yerr=std,
                         ecolor=errorbar_color, capsize=2)
        plt.xlabel('Number of ' + str(SpeciesName) + ' in Single Complex')
        plt.ylabel('Lifetime (s)')
        plt.xticks(ticks=size_list)
        plt.title('Lifetime of Complex')
        if SaveFig:
            plt.savefig('complex_lifetime.png', dpi=500)
        plt.show()
    return size_list, mean, std





import numpy as np
from .PDB_object import ProteinComplex


def PDB_UI():
    """A user friendly UI to allow for open, minupulating, and then outputting .pdb files.

    Functions:

    Open .pdb files:
     - When run, the first thing that will be asked is "Enter pdb file name: []". You must enter the relative / absolute path to the file.
     - Ex: Enter pdb file name: "ioNERDSSPyPi\\TestingFunctions\\1si4.pdb" (Note: I have to add a 2nd '\' before 1, so python does not see it as a weird charecter)
    Chaning distance between interaction sites:
     - After the .pdb file in initilized, it will ask "Would you like to chang...", and ask for you to write 'yes' or 'no'
     - If you write yes, keep reading, if you write no, it will just go to the next section
     - Than it will ask 'which distance' you want to change, and ask for an integer between 0-X.
        - 0: means all distances will be set to the same number you input
        - 1+: That distance will be set to the number inputted. You can find which 'distance' each number refers to by reading and counting down the
        list of Interaction Sites (which is directly above). 1 = the furthest up.
     - Then enter the new distance.
     - Then the initial message will come up again, and repeats this whol process.
    'Normalizing':
     - It will then ask if you want to see the 'default norm vector to (0,0,1)'. Write 'yes' or 'no' if you do or don't
     - It will then ask if you want each molecule's center of mass to be 0,0,0
     - The UI will then spit out the necessary .mol and .inp files to setup a NERDSS simulation

    If you want to make graphs / new .pdb files, you will need to use the 'seperate' commands instead of the UI.
    """
    # naming explanation:
    # variables with word 'total' in the front indicate that it's a list of data for the whole protein
    # variables with word 'split' in the front indicate that it's a list containing n sub-lists and each sub-list contains
    # data for different chains. (n is the number of chains)

    # read in file
    file_name = input("Enter pdb file name: ")
    UI_PDB = ProteinComplex(file_name)

    # user can choose to change the interaction site
    while True:
        answer = input(
            "Would you like to change the distance between interaction site (Type 'yes' or 'no'): "
        )
        if answer == "no":
            print("Calculation is completed.")
            break
        if answer == "yes":
            while True:
                n = int(
                    input(
                        "Which distance would you like to change (please enter an integer no greater than %.0f or enter 0 to set all distance to a specific number): "
                        % (len(UI_PDB.int_site_distance))
                    )
                )
                if n in range(0, len(UI_PDB.int_site_distance) + 1):
                    while True:
                        new_distance = float(input("Please enter new distance: "))
                        # decreasing distance & increasing distance
                        if new_distance >= 0:
                            UI_PDB.change_sigma(
                                ChangeSigma=True, SiteList=[n], NewSigma=[new_distance]
                            )
                            break
                        else:
                            print("Invalid number, please try again.")
                            break
                    break
                else:
                    print("Invalid answer, please try again.")
                    break
        else:
            print("Invalid answer, please try again.")

    # normalize vector
    UI_PDB.calc_angle()

    # ask to whether display 3D graph
    while True:
        answer3 = input(
            "Display a 3D plot of the protein complex? (Type 'yes' or 'no'): "
        )
        if answer3 == "yes":
            UI_PDB.plot_3D()
            break
        if answer3 == "no":
            break
        else:
            print("Invalid answer, please try again.")

    # asking whether to center the COM of every chain to origin.
    while True:
        answer2 = input(
            "Do you want each chain to be centered at center of mass? (Type 'yes' or 'no'): "
        )
        if answer2 == "yes":
            UI_PDB.norm_COM()
            break
        if answer2 == "no":
            break
        else:
            print("Invalid answer, please try again.")

    # writing parameters into a file
    UI_PDB.write_input()
    return 0



from ..locate_pos.nerdss_PDB.read_PDB import read_PDB
import matplotlib.pyplot as plt

def plot_proteins(FileName: str):
    """Create a 3D plot that shows loation of each protein section
    """

    #reads in the PDB file. 
    # site_array: list that holds information about every protein site. 
    # site_dict: dictionary so you can index for site_array for specific attributes.
    # num_name_dict: holds the protein type for every protein number
    # main_pdb_list: literally the pdb file read in. Every line = pdb file
    site_array,site_dict,num_name_dict,main_pdb_list = read_PDB(file_name=FileName,drop_COM=False)
    
    #create plot
    fig = plt.figure(figsize=[12.8,9.6])
    ax = fig.add_subplot(projection="3d")

    for site in site_array:
        if site[site_dict["Site_Name"]] == "COM":
            #get x,y,z
            x = site[site_dict["x_coord"]]
            y = site[site_dict["y_coord"]]
            z = site[site_dict["z_coord"]]
        
            if site[site_dict["Protein_Name"]] == "cla":
                ax.scatter(x,y,z,marker="o")
            elif site[site_dict["Protein_Name"]] == "ap2":
                ax.scatter(x,y,z,marker="X")
            else:
                ax.scatter(x,y,z,marker="H")

    
    ax.set_xlabel("x_coord")
    ax.set_ylabel("y_coord")
    ax.set_zlabel("z_coord")

    plt.show()



from .cube_face_leg_reduce_coord_gen import cube_face_leg_reduce_coord_gen
import numpy as np

def cube_face_input_coord(radius: float, sigma: float):
    """Generates input coordinates for a cube face simulation.

    This function generates input coordinates for a cube face simulation, based on the radius and sigma values
    provided. The calculation is performed using the `cube_face_leg_reduce_coord_gen` function from the
    `.cube_face_leg_reduce_coord_gen` module to obtain reduced coordinates of the cube face, and then
    performs various calculations to derive the input coordinates.

    Args:
        radius (float): The radius of the cube.
        sigma (float): The sigma value for the simulation.

    Returns:
        List: Contains the input coordinates for the cube face simulation, in the following order:
        [COM, lg1, lg2, lg3, lg4, n], where COM is the center of mass of the cube face, lg1, lg2, lg3, and lg4 are
        the leg vectors of the cube face, and n is a vector pointing towards the center of the cube face.


    Example:
        >>> cube_face_input_coord(1.0, 0.1)
        [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [1.0, 0.0, 0.0], [0.0, -1.0, 0.0], [-0.0, -0.0, -0.0]]
    """
    
    coor = cube_face_leg_reduce_coord_gen(radius, sigma)
    coor_ = np.array(coor[0])
    COM = np.around(coor_[0] - coor_[0], 7)
    lg1 = coor_[1] - coor_[0]
    lg2 = coor_[2] - coor_[0]
    lg3 = coor_[3] - coor_[0]
    lg4 = coor_[4] - coor_[0]
    n = -coor_[0]
    return [COM, lg1, lg2, lg3, lg4, n]





from ..gen_platonic.mid_pt import mid_pt
from .octa_face_COM_coord import octa_face_COM_coord


def octa_face_COM_leg_coord(a: float, b: float, c: float):
    """Calculate the coordinates of the center of mass (COM) and midpoints of the legs
    of an octahedron face.

    Given the coordinates of three vertices of an octahedron face (a, b, c), this function
    calculates the coordinates of the center of mass (COM) and the midpoints of the legs of
    that face using the 'octa_face_COM_coord' and 'mid_pt' functions from the respective
    modules.

    Args:
        a (float): The coordinates of the first vertex of the octahedron face as a list or tuple
                   of three floats representing the x, y, and z coordinates, respectively.
        b (float): The coordinates of the second vertex of the octahedron face as a list or tuple
                   of three floats representing the x, y, and z coordinates, respectively.
        c (float): The coordinates of the third vertex of the octahedron face as a list or tuple
                   of three floats representing the x, y, and z coordinates, respectively.

    Returns:
        list: A list of four elements:
            - A list of three floats representing the x, y, and z coordinates of the center of mass (COM)
              for the octahedron face.
            - A list of three floats representing the x, y, and z coordinates of the midpoint of the leg
              connecting vertices a and b.
            - A list of three floats representing the x, y, and z coordinates of the midpoint of the leg
              connecting vertices b and c.
            - A list of three floats representing the x, y, and z coordinates of the midpoint of the leg
              connecting vertices c and a.

    Example:
        To calculate the coordinates of the center of mass and midpoints of the legs for an octahedron
        face with vertices a = [1.0, 2.0, 3.0], b = [4.0, 5.0, 6.0], and c = [7.0, 8.0, 9.0]:
        >>> octa_face_COM_leg_coord([1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0])
        [[3.5, 4.5, 5.5], [2.5, 3.5, 4.5], [5.5, 6.5, 7.5], [4.0, 5.0, 6.0]]

    """
    COM_leg = []
    COM_leg.append(octa_face_COM_coord(a, b, c))
    COM_leg.append(mid_pt(a, b))
    COM_leg.append(mid_pt(b, c))
    COM_leg.append(mid_pt(c, a))
    return COM_leg





from .dode_vert_COM_leg_gen import dode_vert_COM_leg_gen
from .dode_vert_leg_reduce import dode_vert_leg_reduce

def dode_vert_leg_reduce_coor_gen(radius: float, sigma: float):
    """
    Generates reduced center of mass (COM) and leg vectors for a dodecahedron vertex based on radius and sigma values.

    This function generates a list of reduced center of mass (COM) and leg vectors for a dodecahedron vertex based on the
    provided radius and sigma values. The reduced COM and leg vectors are obtained by calling the dode_vert_COM_leg_gen
    function to generate the original COM and leg vectors, and then passing them to the dode_vert_leg_reduce function to
    reduce their lengths. The reduced COM and leg vectors are stored in a list and returned.

    Args:
        radius (float): The radius of the dodecahedron.
        sigma (float): The sigma value for the dodecahedron vertex.

    Returns:
        list: A list of lists, where each inner list contains the reduced center of mass (COM) and leg vectors for a
        dodecahedron vertex.
    """

    COM_leg_list = dode_vert_COM_leg_gen(radius)
    COM_leg_red_list = []
    for elements in COM_leg_list:
        temp_list = []
        temp_list.append(elements[0])
        i = 1
        while i <= 3:
            temp_list.append(dode_vert_leg_reduce(
                elements[0], elements[i], sigma))
            i += 1
        COM_leg_red_list.append(temp_list)
    return COM_leg_red_list





import math
import numpy as np


def fitSphere(x, y, z):
    """This function takes three arrays of equal length x, y and z and returns the radius and center of a sphere that best fits the data.

    Parameters
        x: array of x coordinates
        y: array of y coordinates
        z: array of z coordinates
    Returns
        A tuple of four elements:
            radius: the radius of the best fit sphere
            x_center: the x coordinate of the sphere's center
            y_center: the y coordinate of the sphere's center
            z_center: the z coordinate of the sphere's center
    """
    A = np.zeros((len(x), 4))
    A[:, 0] = 2*x
    A[:, 1] = 2*y
    A[:, 2] = 2*z
    A[:, 3] = 1
    f = np.zeros((len(x), 1))
    f[:, 0] = x*x+y*y+z*z
    C, residules, rank, singval = np.linalg.lstsq(A, f, rcond=None)
    t = (C[0]*C[0])+(C[1]*C[1])+(C[2]*C[2])+C[3]
    radius = math.sqrt(t)
    return radius, C[0], C[1], C[2]





from setuptools import setup, find_packages

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Science/Research",
    "Operating System :: OS Independent",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3",
]

setup(
    name="ioNERDSS",
    version="1.0.30",
    requires=["python (>=3.6)"],
    description="Package for analysing NERDSS inputs and outputs.",
    long_description="A python package for analysing inputs and outputs for NERDSS simulator, including generating input files for Platonic solids and output visualization, etc.",
    url="",
    author="Zixiu (Hugh) Liu",
    author_email="zliu140@jhu.edu",
    license="MIT",
    classifiers=classifiers,
    keywords="NERDSS Simulation, Analysis Tools",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "pandas",
        "matplotlib",
        "seaborn",
        "tqdm",
        "PyQt6",
        "pyqtgraph",
        "PyOpenGL",
        "biopython",
    ],
)



from .dode_vert_write import dode_vert_write


def dode_vert(radius: float, sigma: float):
    """
    Generates and writes vertex coordinates for a dodecahedron to a file.

    Args:
        radius (float): Radius of the dodecahedron.
        sigma (float): Sigma value for generating vertex coordinates.

    Returns:
        parm.inp/cube.mol file: inputs for NERDSS
    """
    
    dode_vert_write(radius, sigma)
    print('File writing complete!')
    return 0





# Form implementation generated from reading ui file 'plot_copy_num.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_PlotCopyNum(object):
    def setupUi(self, PlotCopyNum):
        PlotCopyNum.setObjectName("PlotCopyNum")
        PlotCopyNum.resize(640, 480)
        self.listWidgetSpecies = QtWidgets.QListWidget(parent=PlotCopyNum)
        self.listWidgetSpecies.setGeometry(QtCore.QRect(20, 10, 91, 461))
        self.listWidgetSpecies.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.MultiSelection)
        self.listWidgetSpecies.setObjectName("listWidgetSpecies")
        self.radioButtonPlotSep = QtWidgets.QRadioButton(parent=PlotCopyNum)
        self.radioButtonPlotSep.setGeometry(QtCore.QRect(140, 10, 161, 21))
        self.radioButtonPlotSep.setObjectName("radioButtonPlotSep")
        self.radioButtonPlotSum = QtWidgets.QRadioButton(parent=PlotCopyNum)
        self.radioButtonPlotSum.setGeometry(QtCore.QRect(320, 10, 161, 21))
        self.radioButtonPlotSum.setObjectName("radioButtonPlotSum")
        self.pushButtonPlotSpecies = QtWidgets.QPushButton(parent=PlotCopyNum)
        self.pushButtonPlotSpecies.setGeometry(QtCore.QRect(450, 10, 80, 23))
        self.pushButtonPlotSpecies.setObjectName("pushButtonPlotSpecies")
        self.pushButtonSaveSpecies = QtWidgets.QPushButton(parent=PlotCopyNum)
        self.pushButtonSaveSpecies.setGeometry(QtCore.QRect(310, 450, 131, 21))
        self.pushButtonSaveSpecies.setObjectName("pushButtonSaveSpecies")
        self.widget = MPLWidget(parent=PlotCopyNum)
        self.widget.setGeometry(QtCore.QRect(120, 40, 511, 401))
        self.widget.setObjectName("widget")

        self.retranslateUi(PlotCopyNum)
        QtCore.QMetaObject.connectSlotsByName(PlotCopyNum)

    def retranslateUi(self, PlotCopyNum):
        _translate = QtCore.QCoreApplication.translate
        PlotCopyNum.setWindowTitle(_translate("PlotCopyNum", "Plot Copy Number"))
        self.radioButtonPlotSep.setText(_translate("PlotCopyNum", "Plot Separately"))
        self.radioButtonPlotSum.setText(_translate("PlotCopyNum", "Plot Sum"))
        self.pushButtonPlotSpecies.setText(_translate("PlotCopyNum", "Plot"))
        self.pushButtonSaveSpecies.setText(_translate("PlotCopyNum", "Save to CSV"))
from .mplwidget import MPLWidget


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    PlotCopyNum = QtWidgets.QWidget()
    ui = Ui_PlotCopyNum()
    ui.setupUi(PlotCopyNum)
    PlotCopyNum.show()
    sys.exit(app.exec())



import numpy as np


def read_transition_matrix(FileName: str, SpeciesName: str, InitialTime: float, FinalTime: float):
    """
    Parses transition_matrix_time.dat, and returns the matrices at two given time points.

    Args:
        FileName (str): The name of the file to be read.
        SpeciesName (str): The name of the species to be analyzed.
        InitialTime (float): The initial time point of interest.
        FinalTime (float): The final time point of interest.

    Returns:
          A tuple that contains:
           - NumPy array of the transition matrix at the initial time point
           - NumPy array of the transition matrix at the final time point
    """
    ti_matrix = []
    tf_matrix = []
    speciesDistance = 0

    #read the file (slowest part I promise)
    with open(FileName, 'r') as file:
        lines = file.readlines()

    #determine distance between times/speciesname
    for index,line in enumerate(lines[1:]):

        #distance b/w time line and the correct species name line
        if speciesDistance == 0 and line[0:4] == SpeciesName:
            speciesDistance = index + 1

        #distance between 2 different time slines
        if line[0:5] == 'time:': 
            distance = index + 1
            break
    
    #if no species was found, raise error
    if speciesDistance == 0:
        raise Exception("No species name found")


    #go through each 'time' line
    for index,line in enumerate(lines[::distance]):
            
        #checks if it equal to the initial time
        if float(line.split(' ')[1]) == InitialTime:
            index_real = (distance*index) + 1
            index_start_read = index_real + speciesDistance

            #get data from this time's matrix, and put it into an array (reads until it hits lifetime)
            for data in lines[index_start_read:-1]:
                if data[0:8] != 'lifetime':
                    info = data.strip(' ').strip('\n').split(' ')
                    temp_list = []
                    for value in info:
                        temp_list.append(int(value))
                    ti_matrix.append(temp_list)
                else:
                    break


        #checks if it equal to the final time
        if float(line.split(' ')[1]) == FinalTime:
            index_real = (distance*index) + 1
            index_start_read = index_real + 2
            
            #get data from this time's matrix, and put it into an array (reads until it hits lifetime
            for data in lines[index_start_read:-1]:
                if data[0:8] != 'lifetime':
                    info = data.strip(' ').strip('\n').split(' ')
                    temp_list = []
                    for value in info:
                        temp_list.append(int(value))
                    tf_matrix.append(temp_list)
                else:
                    break
            
    #if no matrices found
    if ti_matrix == []:
        raise Exception('Time initial not found')
    if tf_matrix == []:
        raise Exception('Time final not found')        
    
    #output
    ti_matrix = np.array(ti_matrix)
    tf_matrix = np.array(tf_matrix)
    return ti_matrix, tf_matrix


from .icos_vert_input_coord import icos_vert_input_coord
from .icos_vert_center_coor import icos_vert_center_coor
from .icos_vert_check_dis import icos_vert_check_dis
from .icos_vert_norm_input import icos_vert_norm_input


def icos_vert_write(radius: float, sigma: float):
    """Write input file for icosahedron vertex-centered simulation.

    This function writes an input file in the required format for a icosahedron vertex-centered
    simulation. The input file contains parameters, boundaries, molecules, and reactions
    for the simulation.

    Args:
        radius (float): Radius of the icosahedron.
        sigma (float): Sigma value for the simulation.

    Returns:
        parm.inp/icos.mol: input files for NERDSS


    Raises:
        None

    Examples:
        >>> icos_vert_write(5.0, 0.5)

    Notes:
        - The input file is written to a file named 'parm.inp' in the current directory.
    """
    COM_, lg1_, lg2_, lg3_, lg4_, lg5_, n_ = icos_vert_input_coord(
        radius, sigma)
    cen_ = icos_vert_center_coor(lg1_, lg2_, lg3_, lg4_, lg5_)
    scaler, dis_ = icos_vert_check_dis(
        cen_, COM_, lg1_, lg2_, lg3_, lg4_, lg5_)
    COM, lg1, lg2, lg3, lg4, lg5, n = icos_vert_norm_input(scaler, dis_)

    f = open('parm.inp', 'w')
    f.write(' # Input file (icosahedron vertex-centered)\n\n')
    f.write('start parameters\n')
    f.write('    nItr = 10000000 #iterations\n')
    f.write('    timeStep = 0.1\n')
    f.write('    timeWrite = 10000\n')
    f.write('    pdbWrite = 10000\n')
    f.write('    trajWrite = 10000\n')
    f.write('    restartWrite = 50000\n')
    f.write('    checkPoint = 1000000\n')
    f.write('    overlapSepLimit = 7.0\n')
    f.write('end parameters\n\n')
    f.write('start boundaries\n')
    f.write('    WaterBox = [500,500,500]\n')
    f.write('end boundaries\n\n')
    f.write('start molecules\n')
    f.write('    icos : 200\n')
    f.write('end molecules\n\n')
    f.write('start reactions\n')
    f.write('    icos(lg1) + icos(lg1) <-> icos(lg1!1).icos(lg1!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg2) + icos(lg2) <-> icos(lg2!1).icos(lg2!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg3) + icos(lg3) <-> icos(lg3!1).icos(lg3!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg4) + icos(lg4) <-> icos(lg4!1).icos(lg4!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg5) + icos(lg5) <-> icos(lg5!1).icos(lg5!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg1) + icos(lg2) <-> icos(lg1!1).icos(lg2!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg1) + icos(lg3) <-> icos(lg1!1).icos(lg3!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg1) + icos(lg4) <-> icos(lg1!1).icos(lg4!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg1) + icos(lg5) <-> icos(lg1!1).icos(lg5!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg2) + icos(lg3) <-> icos(lg2!1).icos(lg3!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg2) + icos(lg4) <-> icos(lg2!1).icos(lg4!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg2) + icos(lg5) <-> icos(lg2!1).icos(lg5!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg3) + icos(lg4) <-> icos(lg3!1).icos(lg4!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg3) + icos(lg5) <-> icos(lg3!1).icos(lg5!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    icos(lg4) + icos(lg5) <-> icos(lg4!1).icos(lg5!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('end reactions\n')

    f = open('icos.mol', 'w')
    f.write('##\n')
    f.write('# Icosahedron (vertex-centered) information file.\n')
    f.write('##\n\n')
    f.write('Name = icos\n')
    f.write('checkOverlap = true\n\n')
    f.write('# translational diffusion constants\n')
    f.write('D = [13.0, 13.0, 13.0]\n\n')
    f.write('# rotational diffusion constants\n')
    f.write('Dr = [0.03, 0.03, 0.03]\n\n')
    f.write('# Coordinates\n')
    f.write('COM   ' + str(round(COM[0], 8)) + '   ' +
            str(round(COM[1], 8)) + '   ' + str(round(COM[2], 8)) + '\n')
    f.write('lg1   ' + str(round(lg1[0], 8)) + '   ' +
            str(round(lg1[1], 8)) + '   ' + str(round(lg1[2], 8)) + '\n')
    f.write('lg2   ' + str(round(lg2[0], 8)) + '   ' +
            str(round(lg2[1], 8)) + '   ' + str(round(lg2[2], 8)) + '\n')
    f.write('lg3   ' + str(round(lg3[0], 8)) + '   ' +
            str(round(lg3[1], 8)) + '   ' + str(round(lg3[2], 8)) + '\n')
    f.write('lg4   ' + str(round(lg4[0], 8)) + '   ' +
            str(round(lg4[1], 8)) + '   ' + str(round(lg4[2], 8)) + '\n')
    f.write('lg5   ' + str(round(lg5[0], 8)) + '   ' +
            str(round(lg5[1], 8)) + '   ' + str(round(lg5[2], 8)) + '\n')
    f.write('\n')
    f.write('# bonds\n')
    f.write('bonds = 5\n')
    f.write('com lg1\n')
    f.write('com lg2\n')
    f.write('com lg3\n')
    f.write('com lg4\n')
    f.write('com lg5\n')
    f.write('\n')


# OCTAHEDRON FACE AS COM




from .cube_vert_coord import cube_vert_coord
from .cube_vert_COM_leg import cube_vert_COM_leg

def cube_vert_COM_leg_gen(radius: float):
    """Generates a list of midpoints of line segments between a central point and other points on a cube.

    This function calculates the midpoints of line segments between a central point and other points on a cube, based on the given
    radius.

    Args:
        radius (float): The radius of the cube.

    Returns:
        list: A list containing eight sub-lists, each containing four floating-point values rounded to 10 decimal places,
        representing the central point and the midpoints of line segments between the central point and other points on the cube.

    Example:
        cube_vert_COM_leg_gen(1.0)
        # Generates a list of midpoints of line segments between the central point and other points on a cube with a radius of 1.0.
        # The list contains eight sub-lists, each containing four floating-point values rounded to 10 decimal places.
    """

    coord = cube_vert_coord(radius)
    COM_leg_list = []
    COM_leg_list.append(cube_vert_COM_leg(
        coord[0], coord[1], coord[2], coord[3]))
    COM_leg_list.append(cube_vert_COM_leg(
        coord[1], coord[0], coord[4], coord[6]))
    COM_leg_list.append(cube_vert_COM_leg(
        coord[2], coord[0], coord[4], coord[5]))
    COM_leg_list.append(cube_vert_COM_leg(
        coord[3], coord[0], coord[5], coord[6]))
    COM_leg_list.append(cube_vert_COM_leg(
        coord[4], coord[1], coord[2], coord[7]))
    COM_leg_list.append(cube_vert_COM_leg(
        coord[5], coord[2], coord[3], coord[7]))
    COM_leg_list.append(cube_vert_COM_leg(
        coord[6], coord[1], coord[3], coord[7]))
    COM_leg_list.append(cube_vert_COM_leg(
        coord[7], coord[4], coord[5], coord[6]))
    return COM_leg_list





from ..gen_platonic.mid_pt import mid_pt
from .icos_face_COM_coord import icos_face_COM_coord


def icos_face_COM_leg_coord(a: float, b: float, c: float):
    """Calculate the center of mass (COM) leg coordinates for an icosahedron face.

    This function calculates the center of mass (COM) leg coordinates for an icosahedron face
    with three given coordinates `a`, `b`, and `c` using the `icos_face_COM_coord()` and `mid_pt()`
    functions from the `icos_face_COM_coord` and `mid_pt` modules respectively. The COM leg coordinates
    are calculated as follows:
    - The COM of the face using `icos_face_COM_coord()` function
    - The mid-point of each pair of vertices using `mid_pt()` function

    Args:
        a (float): The first coordinate of the icosahedron face.
        b (float): The second coordinate of the icosahedron face.
        c (float): The third coordinate of the icosahedron face.

    Returns:
        List[list[float]]: The center of mass (COM) leg coordinates as a list of lists of three floats.
            The list has four sub-lists, each containing the COM leg coordinates for a pair of vertices.

    Examples:
        >>> a = [0, 0, 0]
        >>> b = [1, 1, 1]
        >>> c = [2, 2, 2]
        >>> icos_face_COM_leg_coord(a, b, c)
        [[1.3660254037847, 1.3660254037847, 1.3660254037847],
         [0.5, 0.5, 0.5],
         [1.5, 1.5, 1.5],
         [1.0, 1.0, 1.0]]

    Notes:
        - The COM leg coordinates are calculated using the `icos_face_COM_coord()` function for the face
          and `mid_pt()` function for the mid-points of pairs of vertices.
        - The calculated COM leg coordinates are returned as a list of lists, where each sub-list contains
          three floats representing the COM leg coordinates for a pair of vertices.
    """
    COM_leg = []
    COM_leg.append(icos_face_COM_coord(a, b, c))
    COM_leg.append(mid_pt(a, b))
    COM_leg.append(mid_pt(b, c))
    COM_leg.append(mid_pt(c, a))
    return COM_leg





import numpy as np
import copy

def hist_to_csv(FullHist: list, SpeciesList: list, OpName: str, Single: bool = False):
    """Creates a .csv (spreadsheet) file from a histogram.dat file (multi-species). If given multiple histograms, it will calculate the average between them.

    Args:
        FullHist (list): holds all of the histogram data
        SpeciesList (list): list of each included species type
        OpName (str): name of the outputted .csv file
        Single (bool): whether or not it is a single histogram

    Returns:
        histogram.csv file: Each row is a different time stamp (all times listed in column A). Each column is a different size of complex molecule (all sizes listed in row 1). Each box 
        is the number of that complex molecule at that time stamp.

    wow this is inefficient, but.... i dont care :D
    """
    
    column_list = [] #holds the name of each column (Time + each complex name)
    time_list = [] #holds each time. Index corresponds to a sublist in name_count_dict_List
    name_count_dict_list = [] #holds each name/count. Index corresponds with size_list
    FakeHist = copy.deepcopy(FullHist)

    #turn single histogram file format into multi histogram
    if Single:
        for file_index,file in enumerate(FakeHist):
            for time_index,time in enumerate(file):
                temp_list = []
                for count_index,count in enumerate(time[1]):
                    sub_temp_list = []
                    sub_temp_list.append(time[2][count_index])
                    sub_temp_list.append(count)
                    temp_list.append(sub_temp_list)
                FakeHist[file_index][time_index] = [FakeHist[file_index][time_index][0]] + temp_list


    #determine longest file
    length = 0
    for file_index,file in enumerate(FakeHist):
        if len(file) > length:
            length = len(file)
            length_index = file_index 
    

    #create list with dictionaries for each timestep
    for na in range(0,length+1):
        name_count_dict_list.append({})

    #goes through every timestep (usese the longest file for this)
    for time_index,time in enumerate(FakeHist[file_index]):

        #create list of every timestep
        time_list.append(time[0])

        #go through every file
        counter = 0
        for file_index,file in enumerate(FakeHist):

            #go through every complex in this file if this timestep exists
            if len(file) > time_index:
                counter = counter + 1
                for complexes in file[time_index][1:]:

                    #get name
                    name = ""
                    for index_sp,species in enumerate(complexes[:-1]):
                        name = f"{name}{SpeciesList[index_sp]}: {species}.  "
                        if name not in name_count_dict_list[time_index].keys():
                            name_count_dict_list[time_index][name] = []
                    
                    #get count
                    name_count_dict_list[time_index][name].append(complexes[-1])

                    #creates a list of every 'name'
                    if name not in column_list:
                        column_list.append(name)

        #determine the average
        for key,value in name_count_dict_list[time_index].items():
            #ensure the list has all of the included histograms
            if len(value) < counter:
                for na in range(len(value), counter):
                    value.append(0)

            #takes mean of list
            name_count_dict_list[time_index][key] = np.mean(value)


    #write the file!
    with open(f'{OpName}.csv', 'w') as write_file:
        
        #create column names
        head = 'Time(s):'
        for column in column_list:
            head += ','
            head += column
        head += '\n'
        write_file.write(head)

        #write the bulk of the file
        for index,timestep in enumerate(time_list):
            
            #initilize writing
            write = ''

            #write time to string
            write += f"{str(timestep)}"

            #write data to string
            for column in column_list:
                write += ','
                if column in name_count_dict_list[index].keys():
                    write += str(name_count_dict_list[index][column])
                else:
                    write += '0'
            
            #commit to file
            write += '\n'
            write_file.write(write)






from .tetr_vert_leg_reduce_coor_gen import tetr_vert_leg_reduce_coor_gen
import numpy as np


def tetr_vert_input_coord(radius: float, sigma: float):
    """Generate the input coordinates for a regular tetrahedron given the radius and sigma.

    Args:
        radius (float): The radius of the circumsphere of the tetrahedron.
        sigma (float): The scaling factor for the coordinates.

    Returns:
        tuple: A tuple containing the center of mass (COM) and three leg vectors of the tetrahedron,
        as well as the normalized vector of the first vertex.

    Example:
        >>> tetr_vert_input_coord(1.0, 0.5)
        (array([0., 0., 0.]), array([0.5, 0., 0.]), array([0., 0.5, 0.]), array([0., 0., 0.5]), array([1., 0., 0.]))
    """
    coor = tetr_vert_leg_reduce_coor_gen(radius, sigma)
    coor_ = np.array(coor[0])
    COM = np.around(coor_[0] - coor_[0], 8)
    lg1 = np.around(coor_[1] - coor_[0], 8)
    lg2 = np.around(coor_[2] - coor_[0], 8)
    lg3 = np.around(coor_[3] - coor_[0], 8)
    n = np.around(coor_[0]/np.linalg.norm(coor_[0]), 8)
    return COM, lg1, lg2, lg3, n





def octa_vert_coord(radius: float):
    """Calculates the vertex coordinates of an octagon centered at the origin.
    The vertex coordinates are scaled by the given radius.

    Args:
        radius (float): The radius of the octagon.

    Returns:
        list: A list of 6 vertex coordinates, each represented as a list of 3D coordinates [x, y, z].
        The coordinates are scaled by the given radius.

    Example:
        radius = 2.0
        result = octa_vert_coord(radius)
        print(result)
    """
    
    scaler = radius
    v0 = [1, 0, 0]
    v1 = [-1, 0, 0]
    v2 = [0, 1, 0]
    v3 = [0, -1, 0]
    v4 = [0, 0, 1]
    v5 = [0, 0, -1]
    VertCoord = [v0, v1, v2, v3, v4, v5]
    VertCoord_ = []
    for i in VertCoord:
        temp_list = []
        for j in i:
            temp = j*scaler
            temp_list.append(temp)
        VertCoord_.append(temp_list)
    return VertCoord_





import math
import numpy as np


def angle_cal(COM1: float, leg1: float, COM2: float, leg2: float):
    """Calculates angles between vectors based on given inputs.

    Args:
        COM1 (float): Center of Mass (COM) for the first leg.
        leg1 (float): Endpoint of the first leg.
        COM2 (float): Center of Mass (COM) for the second leg.
        leg2 (float): Endpoint of the second leg.

    Returns:
        tuple: A tuple containing the following angles (in radians) rounded to 8 decimal places:
            - theta1 (float): Angle between vector from COM1 to leg1 and vector from leg1 to leg2.
            - theta2 (float): Angle between vector from COM2 to leg2 and vector from leg2 to leg1.
            - phi1 (float): Angle between vectors perpendicular to leg1 and leg2, passing through COM1.
            - phi2 (float): Angle between vectors perpendicular to leg2 and leg1, passing through COM2.
            - omega (float): Angle between vectors perpendicular to leg1 and leg2, passing through leg1 and leg2.
    """
    
    n = 8
    c1 = np.array(COM1)
    p1 = np.array(leg1)
    c2 = np.array(COM2)
    p2 = np.array(leg2)
    v1 = p1 - c1
    v2 = p2 - c2
    sig1 = p1 - p2
    sig2 = -sig1
    theta1 = round(math.acos(np.dot(v1, sig1) /
                   (np.linalg.norm(v1)*np.linalg.norm(sig1))), n)
    theta2 = round(math.acos(np.dot(v2, sig2) /
                   (np.linalg.norm(v2)*np.linalg.norm(sig2))), n)
    t1 = np.cross(v1, sig1)
    t2 = np.cross(v1, c1)  # n1 = c1 here
    t1_hat = t1/np.linalg.norm(t1)
    t2_hat = t2/np.linalg.norm(t2)
    phi1 = round(math.acos(np.around(np.dot(t1_hat, t2_hat), n)), n)
    t3 = np.cross(v2, sig2)
    t4 = np.cross(v2, c2)  # n2 = c2 here
    t3_hat = t3/np.linalg.norm(t3)
    t4_hat = t4/np.linalg.norm(t4)
    phi2 = round(math.acos(np.around(np.dot(t3_hat, t4_hat), n)), n)
    t1_ = np.cross(sig1, v1)
    t2_ = np.cross(sig1, v2)
    t1__hat = t1_/np.linalg.norm(t1_)
    t2__hat = t2_/np.linalg.norm(t2_)
    omega = round(math.acos(np.around(np.dot(t1__hat, t2__hat), n)), n)
    return theta1, theta2, phi1, phi2, omega


# DODECAHEDEON FACE AS COM




import math


def real_PDB_mag(x):
    """
    This function calculates the direct distance b/w 2 coordinates (in a 3d space) based on their dx,dy,dz.

    Args:
        x (list): a list of float values representing the Cartesian coordinates (dx,dy,dz) of the vector.

    Returns:
        A float value representing the magnitude of the vector.

    Example:
        >>> real_PDB_mag([3.0, 4.0, 0.0])
        5.0
    """
    return math.sqrt(sum(i ** 2 for i in x))





import numpy as np


def real_PDB_unit(x: np.ndarray) -> np.ndarray:
    """Returns the unit vector of the input vector x.

    Args:
        x (numpy.ndarray): The input vector.

    Returns:
        numpy.ndarray: The unit vector of x.

    Note:
        If the magnitude of x is zero, this function will return a vector of zeros.
        If the magnitude of x is one, this function will return x.

    Examples:
        >>> real_PDB_unit(np.array([1, 2, 3]))
        array([0.26726124, 0.53452248, 0.80178373])
        >>> real_PDB_unit(np.array([0, 0, 0]))
        array([0, 0, 0])
        >>> real_PDB_unit(np.array([1, 0, 0]))
        array([1, 0, 0])
    """

    x_norm = np.linalg.norm(x)
    if abs(x_norm-1) < 10**-6:
        return x
    elif x_norm < 10**-6:
        return np.zeros(3)
    else:
        return x/x_norm





from .icos_vert_leg_reduce_coor_gen import icos_vert_leg_reduce_coor_gen
import numpy as np


def icos_vert_input_coord(radius: float, sigma: float):
    """Generate input vertex coordinates for an icosahedron with the given radius and sigma.

    The function calculates the input vertex coordinates for an icosahedron with the given radius and sigma,
    using mathematical formulas and numpy operations.

    Args:
        radius (float): The radius of the icosahedron.
        sigma (float): The sigma value for generating the vertex coordinates.

    Returns:
        tuple: A tuple of input vertex coordinates for the icosahedron, including the center of mass (COM),
        and the leg vectors (lg1, lg2, lg3, lg4, lg5) and the normalized normal vector (n).
    """
    coor = icos_vert_leg_reduce_coor_gen(radius, sigma)
    coor_ = np.array(coor[0])
    COM = np.around(coor_[0] - coor_[0], 12)
    lg1 = np.around(coor_[1] - coor_[0], 12)
    lg2 = np.around(coor_[2] - coor_[0], 12)
    lg3 = np.around(coor_[3] - coor_[0], 12)
    lg4 = np.around(coor_[4] - coor_[0], 12)
    lg5 = np.around(coor_[5] - coor_[0], 12)
    n = np.around(coor_[0]/np.linalg.norm(coor_[0]), 12)
    return COM, lg1, lg2, lg3, lg4, lg5, n





import numpy as np
from ..gen_platonic.mid_pt import mid_pt


def icos_vert_COM_leg(COM: float, a: float, b: float, c: float, d: float, e: float):
    """Calculate center of mass (COM) and legs from COM to each point.

    Args:
        COM (float): The center of mass point.
        a (float): Point A.
        b (float): Point B.
        c (float): Point C.
        d (float): Point D.
        e (float): Point E.

    Returns:
        list: A list containing the center of mass and legs coordinates, rounded to 10 decimal places.
    """
    lega = mid_pt(COM, a)
    legb = mid_pt(COM, b)
    legc = mid_pt(COM, c)
    legd = mid_pt(COM, d)
    lege = mid_pt(COM, e)
    result = [np.around(COM, 10), np.around(lega, 10), np.around(
        legb, 10), np. around(legc, 10), np.around(legd, 10), np.around(lege, 10)]
    return result





def dtb_PDB_write_input(Result: tuple):
    """
    Writes '.inp' and '.mol' files based on the calculations and modifications performed by the previous functions. Multiple '.mol' files and a single '.inp' file will be created in the same directory as the Jupyter Notebook file once the function finishes running.

    Parameters:
        Result (9 length tuple): The output result of function(s): 'angle','COM'

    Returns:
        File with new inputs for NERDSS.
    """

    (
        reaction_chain,
        new_int_site,
        new_int_site_distance,
        unique_chain,
        COM,
        angle,
        normal_point_lst1,
        normal_point_lst2,
        one_site_chain,
    ) = Result
    f = open("parm.inp", "w")
    f.write(" # Input file\n\n")
    f.write("start parameters\n")
    f.write("    nItr = 10000\n")
    f.write("    timestep = 0.1\n")
    f.write("    timeWrite = 500\n")
    f.write("    trajWrite = 500\n")
    f.write("    pdbWrite = 500\n")
    f.write("    restartWrite = 5000\n")
    f.write("end parameters\n\n")
    f.write("start boundaries\n")
    f.write("    WaterBox = [494,494,494] #nm\n")
    f.write("end boundaries\n\n")
    f.write("start molecules\n")
    for i in range(len(unique_chain)):
        f.write("     %s:10\n" % (unique_chain[i]))
    f.write("end molecules\n\n")
    f.write("start reactions\n")
    for i in range(len(reaction_chain)):
        molecule1_lower = reaction_chain[i][0].lower()
        molecule2_lower = reaction_chain[i][1].lower()
        f.write(
            "    #### %s - %s ####\n" % (reaction_chain[i][0], reaction_chain[i][1])
        )
        f.write(
            "    %s(%s) + %s(%s) <-> %s(%s!1).%s(%s!1)\n"
            % (
                reaction_chain[i][0],
                molecule2_lower,
                reaction_chain[i][1],
                molecule1_lower,
                reaction_chain[i][0],
                molecule2_lower,
                reaction_chain[i][1],
                molecule1_lower,
            )
        )
        f.write("    onRate3Dka = 10\n")
        f.write("    offRatekb = 1\n")
        f.write("    sigma = %f\n" % angle[i][5])
        f.write(
            "    norm1 = [%.6f,%.6f,%.6f]\n"
            % (
                normal_point_lst1[i][0],
                normal_point_lst1[i][1],
                normal_point_lst1[i][2],
            )
        )
        f.write(
            "    norm2 = [%.6f,%.6f,%.6f]\n"
            % (
                normal_point_lst2[i][0],
                normal_point_lst2[i][1],
                normal_point_lst2[i][2],
            )
        )
        if reaction_chain[i][0] in one_site_chain:
            angle[i][2] = "nan"
        if reaction_chain[i][1] in one_site_chain:
            angle[i][3] = "nan"
        f.write(
            "    assocAngles = ["
            + str(angle[i][0])
            + ","
            + str(angle[i][1])
            + ","
            + str(angle[i][2])
            + ","
            + str(angle[i][3])
            + ","
            + str(angle[i][4])
            + "]\n\n"
        )
    f.write("end reactions")
    f.close()

    for i in range(len(unique_chain)):
        mol_file = str(unique_chain[i]) + ".mol"
        f = open(mol_file, "w")
        f.write("##\n# %s molecule information file\n##\n\n" % unique_chain[i])
        f.write("Name    = %s\n" % unique_chain[i])
        f.write("checkOverlap = true\n\n")
        f.write("# translational diffusion constants\n")
        f.write("D       = [12.0,12.0,12.0]\n\n")
        f.write("# rotational diffusion constants\n")
        f.write("Dr      = [0.5,0.5,0.5]\n\n")
        f.write("# Coordinates, with states below, or\n")
        f.write("COM     %.4f    %.4f    %.4f\n" % (COM[i][0], COM[i][1], COM[i][2]))
        reaction_chain_merged = []
        chain_string = []
        bond_counter = 0
        for a in range(len(reaction_chain)):
            for b in range(2):
                reaction_chain_merged.append(reaction_chain[a][b])
        if unique_chain[i] not in reaction_chain_merged:
            break
        if unique_chain[i] in reaction_chain_merged:
            bond_counter = 0
            for m in range(len(reaction_chain)):
                if unique_chain[i] == reaction_chain[m][0]:
                    bond_counter += 1
                    chain_name = str(reaction_chain[m][1])
                    chain_string.append(chain_name.lower())
                    f.write(
                        "%s       %.4f    %.4f    %.4f\n"
                        % (
                            chain_name.lower(),
                            new_int_site[m][0][0],
                            new_int_site[m][0][1],
                            new_int_site[m][0][2],
                        )
                    )
                elif unique_chain[i] == reaction_chain[m][1]:
                    bond_counter += 1
                    chain_name = str(reaction_chain[m][0])
                    f.write(
                        "%s       %.4f    %.4f    %.4f\n"
                        % (
                            chain_name.lower(),
                            new_int_site[m][1][0],
                            new_int_site[m][1][1],
                            new_int_site[m][1][2],
                        )
                    )
                    chain_string.append(chain_name)
        f.write("\nbonds = %d\n" % bond_counter)
        for j in range(bond_counter):
            f.write("COM %s\n" % chain_string[j])
    print("Input files written complete.")
    return 0



from .octa_face_vert_coord import octa_face_vert_coord
from .octa_face_COM_leg_coord import octa_face_COM_leg_coord


def octa_face_COM_leg_list_gen(radius: float):
    """Generate a list of center of mass (COM) and midpoints of legs for all octahedron faces.

    Given the radius of an octahedron, this function generates a list of center of mass (COM)
    and midpoints of legs for all eight faces of the octahedron using the 'octa_face_vert_coord'
    and 'octa_face_COM_leg_coord' functions from the respective modules.

    Args:
        radius (float): The radius of the octahedron.

    Returns:
        list: A list of eight elements, each element containing a list of four sub-elements:
            - A list of three floats representing the x, y, and z coordinates of the center of mass (COM)
              for a particular octahedron face.
            - A list of three floats representing the x, y, and z coordinates of the midpoint of the leg
              connecting vertices of that face.
            - A list of three floats representing the x, y, and z coordinates of the midpoint of the leg
              connecting vertices of that face.
            - A list of three floats representing the x, y, and z coordinates of the midpoint of the leg
              connecting vertices of that face.
    """
    coord = octa_face_vert_coord(radius)
    COM_leg_list = []

    COM_leg_list.append(octa_face_COM_leg_coord(coord[0], coord[2], coord[4]))
    COM_leg_list.append(octa_face_COM_leg_coord(coord[0], coord[3], coord[4]))
    COM_leg_list.append(octa_face_COM_leg_coord(coord[0], coord[3], coord[5]))
    COM_leg_list.append(octa_face_COM_leg_coord(coord[0], coord[2], coord[5]))
    COM_leg_list.append(octa_face_COM_leg_coord(coord[1], coord[2], coord[4]))
    COM_leg_list.append(octa_face_COM_leg_coord(coord[1], coord[3], coord[4]))
    COM_leg_list.append(octa_face_COM_leg_coord(coord[1], coord[3], coord[5]))
    COM_leg_list.append(octa_face_COM_leg_coord(coord[1], coord[2], coord[5]))
    return COM_leg_list





import numpy as np
import matplotlib.pyplot as plt
from ..file_managment.save_vars_to_file import save_vars_to_file


def line_size_over_time(Data: int, full_hist: list, FileNum: int, InitialTime: float, FinalTime: float,
                 SpeciesName: str = "tot", ExcludeSize: int = 0, SpeciesList: list = ["na"], ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
    """Creates a graph counting the number of protein species in a complex molecule over a time period.
        Can either be average count or the max count of that protein species at that time stamp .

    Args:
        Data (int): what will the graph show (1: mean, 2: max)
        full_hist (list): holds all of that data in the histogram.dat file
        FileNum (int): Number of the total input files (file names should be [fileName]_1,[fileName]_2,...)
        InitialTime (float): The starting time. Must not be smaller / larger then times in file.
        FinalTime (float): The ending time. Must not be smaller / larger then times in file.
        SpeciesName (str): The name of the species you want to examine. Should be in the .dat file.
        ExcludeSize (int): Monomers in the complex that are smaller or equal to this number will not be included. 
        SpeciesList (lst, optional): lists all of the species in the doc. Only needed for multi species hists.
        ShowFig (bool, optional): If the plot is shown. Defaults to True.
        SaveFig (bool, optional): If the plot is saved. Defaults to False.
        SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

    Returns:
        graph. X-axis = time. Y-axis = mean number of species in a single complex molecule.
    """
    
    time_list = [] #list of every timestamp
    size_list = [] #list of mean sizes (index of this = index of timestep)
    max_size = 0 #the longest histogram
    max_index = 0 #the index of the longest histogram
    count = 0 #used to find number of species in a complex

    #write name of the plot / y label (if applicable)
    if Data == 1:
        graphTitle = 'Average Number of ' + str(SpeciesName) + ' in Single Complex'
        yLabel = 'Average Number of ' + str(SpeciesName)
    elif Data == 2:
        graphTitle = 'Maximum Number of ' + str(SpeciesName) + ' in Single Complex'
        yLabel = 'Maximum Number of ' + str(SpeciesName)
    else:
        raise Exception("Invalid data type")


    for index,hist in enumerate(full_hist):

        total_size_list = [] #list of every timestamp for this file
        total_time_list = [] #list of mean sizes (index of this = index of timestep)

        #create list of means / timesteps, based on what sizes are excluded/not
        if ExcludeSize >= 0:
            for timestep in hist:
                if InitialTime <= timestep[0] <= FinalTime:
                    total_time_list.append(timestep[0])
                    
                    #create a list of each complex size at this timestamp
                    if SpeciesList != ['na']:
                        timestep_edited = []
                        
                        #if all species are included or just 1 (for multi)
                        if SpeciesName != "tot":
                            ind = SpeciesList.index(SpeciesName)
                            for complex in timestep[1:]:
                                timestep_edited.append(complex[ind])
                        else:
                            for complex in timestep[1:]:
                                count = 0
                                for species in complex[:-1]:
                                    count += species
                                timestep_edited.append(count)
                    else:
                        timestep_edited = timestep[2]
        
                    if ExcludeSize > 0:
                        timestep_edited = [ele for ele in timestep_edited if ele>ExcludeSize] #create new list that only includes elements greater then exclude size
                        if timestep_edited == []: timestep_edited.append(0)

                    if Data == 1: #if it is a mean / max line graph
                        total_size_list.append(np.mean(timestep_edited))
                    else:
                        total_size_list.append(np.max(timestep_edited))
        
            #determine size of hist
            if len(total_size_list) > max_size:
                max_size = len(total_size_list)
                max_index = index


        else:
            print('ExcludeSize cannot smaller than 0!')
            return 0
        
        #add time/size lists to main, cross function lists
        time_list.append(total_time_list)
        size_list.append(total_size_list)

    #ensure all lists are tranposible (have equal shape)
    for index_file,file_size in enumerate(size_list):
        if len(file_size) != max_size:
            zero_list = np.zeros(max_size)
            for index_size,size in enumerate(file_size):
                zero_list[index_size] =+ size
            size_list[index_file] = zero_list


    #transpose list (each sub-list = 1 timesteps across every file)
    size_list_rev = []
    size_list_rev = np.transpose(size_list)

    #find mean and std dev
    mean = []
    std = []

    for index,timestamps in enumerate(size_list_rev):
        
        #if this timestamp is equal to previous, copy previous. 
        if (timestamps == size_list_rev[index-1]).all():
            mean.append(mean[index-1])
            if FileNum > 1: std.append(std[index-1])
        
        #Else calculate new measns/stds
        else:
            mean.append(np.nanmean(timestamps))
            if FileNum > 1: std.append(np.nanstd(timestamps))
    
    #output variables
    if SaveVars:
        if Data == 1:
            save_vars_to_file({"time_stamp":time_list[0],"mean_cmplx_size":mean,"std":std})
        if Data == 2:
            save_vars_to_file({"time_stamp":time_list[0],"max_cmplx_size":mean,"std":std})

    #show figure
    if ShowFig:
        errorbar_color = '#c9e3f6'
        plt.plot(time_list[max_index], mean, color='C0')
        if FileNum > 1:
            plt.errorbar(time_list[max_index], mean, color='C0',
                         yerr=std, ecolor=errorbar_color)
        plt.title(graphTitle)
        plt.xlabel('Time (s)')
        plt.ylabel(yLabel)
        if SaveFig:
            plt.savefig('mean_complex.png', dpi=500)
        plt.show()
    return time_list[0], mean, std





from .read_file import read_file
from .complex_time_3d import complex_time_3d
from ..line_size_over_time import line_size_over_time
from .hist_complex_count import hist_complex_count
from ..hist_to_csv import hist_to_csv
from ..hist_to_df import hist_to_df

class SingleHistogram ():
    """Single Histogram object that holds all data from a histogram.dat to be interpreting in many different ways
    """

    def __init__(self,FileName: str, FileNum: int, InitialTime: float, FinalTime: float, SpeciesName: str):
        """Will initilize the object by reading through the inputted file(s)

        Args:
            FileName (str): file location (relative) histogram.dat that will be read
            FileNum (int): Number of the total input files (file names should be [fileName]_1,[fileName]_2,...)
            InitialTime (float): The starting time. Must not be smaller / larger then times in file.
            FinalTime (float): The ending time. Must not be smaller / larger then times in file.
            SpeciesName (str): The name of the species you want to examine. Should be in the .dat file.        
        """


        #Initilize variables
        self.full_hist = []
        self.FileName = FileName
        self.FileNum = FileNum
        self.InitialTime = InitialTime
        self.FinalTime = FinalTime
        self.SpeciesName = SpeciesName
        
        #setup file naming
        file_name_head = FileName.split('.')[0]
        file_name_tail = FileName.split('.')[1]

        for histogram_file_number in range(1, FileNum+1):
            
            #determining file name (if there are multiple or none)
            if FileNum == 1:
                temp_file_name = FileName
            else:
                temp_file_name = file_name_head + '_' + str(histogram_file_number) + '.' + file_name_tail
            
            #load in the file
            temp_hist = read_file(temp_file_name,SpeciesName)
            self.full_hist.append(temp_hist)
    

    ##Count of monomers in each type / complexes of each type over time (3d)
    def heatmap_complex_count(self, TimeBins: int, xBarSize: int = 1, ShowFig: bool = True,
                        ShowMean: bool = False, ShowStd: bool = False, SaveFig: bool = False, SaveVars: bool = False):
        """Creates a 2D Heatmap from a histogram.dat file that represents the average number of each complex size, over time.

        Args:
            TimeBins (int): The number of bins that the selected time period is divided into.
            xBarSize (int, optional): The size of each data bar in the x-dimension. Defaults to 1.
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            ShowMean (bool, optional): If means will be shown in each box. Defaults to False.
            ShowStd (bool, optional): If std values will be shown in each box. Defaults to False.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.


        Returns:
            2D heatmap. X-axis: size of complex species. Y-axis = time 'bins'. Color = relative number of each species.
        """
        
        return complex_time_3d(1,1,self.full_hist, self.FileNum, self.InitialTime, self.FinalTime,
                self.SpeciesName, TimeBins, xBarSize, ShowFig,
                ShowMean, ShowStd, SaveFig, SaveVars)


    def heatmap_monomer_count(self,TimeBins: int, xBarSize: int = 1, ShowFig: bool = True,
                                    ShowMean: bool = False, ShowStd: bool = False, SaveFig: bool = False, SaveVars: bool = False):
        """Creates a 2D Heatmap from a histogram.dat file that shows average number of monomers in each complex size over time.

        Args:
            TimeBins (int): The number of bins that the selected time period is divided into.
            xBarSize (int, optional): The size of each data bar in the x-dimension. Defaults to 1.
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            ShowMean (bool, optional): If means will be shown in each box. Defaults to False.
            ShowStd (bool, optional): If std values will be shown in each box. Defaults to False.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.


        Returns:
            2D heatmap. X-axis = size of complex species. Y-axis = time. Color = total number of corresponding monomers in N-mers.
        """

        return complex_time_3d(1,2,self.full_hist, self.FileNum, self.InitialTime, self.FinalTime,
                self.SpeciesName, TimeBins, xBarSize, ShowFig,
                ShowMean, ShowStd, SaveFig, SaveVars)      


    def heatmap_monomer_fraction(self,TimeBins: int, xBarSize: int = 1, ShowFig: bool = True,
                                ShowMean: bool = False, ShowStd: bool = False, SaveFig: bool = False, SaveVars: bool = False):
        """Generates a 2D histogram from histogram.dat of the % of the original monomers forming into different complex sizes over time

        Args:
            TimeBins (int): The number of bins that the selected time period is divided into.
            xBarSize (int, optional): The size of each data bar in the x-dimension. Defaults to 1.
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            ShowMean (bool, optional): If means will be shown in each box. Defaults to False.
            ShowStd (bool, optional): If std values will be shown in each box. Defaults to False.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.


        Returns:
            2D heatnao. X-axis = complex species size. Y-axis = time. Color = fraction of monomers forming into that complex at that time
        """
        
        return complex_time_3d(1,3,self.full_hist, self.FileNum, self.InitialTime, self.FinalTime,
                self.SpeciesName, TimeBins, xBarSize, ShowFig,
                ShowMean, ShowStd, SaveFig, SaveVars)
        

    def hist_3d_complex_count(self,TimeBins: int, xBarSize: int = 1, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
        """Takes in a histogram.dat file from NERDSS, and creates a 3D histogram that represents the average count each complex size, over time.

        Args:
            TimeBins (int): The number of bins that the selected time period is divided into.
            xBarSize (int, optional): The size of each data bar in the x-dimension. Defaults to 1.
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.


        Returns:
            Returns a 3D histogram representing the aver count of each complex over time. X-axis = species type/size. Y-axis = averaged time. Z-axis = relative occurance.
        """

        return complex_time_3d(2,1,self.full_hist, self.FileNum, self.InitialTime, self.FinalTime,
                self.SpeciesName, TimeBins, xBarSize, ShowFig,
                False, False, SaveFig, SaveVars)


    def hist_3d_monomer_count(self,TimeBins: int, xBarSize: int = 1, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
        """Takes in a histogram.dat file from NERDSS, and creates a 3D histogram that shows average number of monomers in each complex size over time.

        Args:
            TimeBins (int): The number of bins that the selected time period is divided into.
            xBarSize (int, optional): The size of each data bar in the x-dimension. Defaults to 1.
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.


        Returns:
            Returns a 3D histogram representing the aver count of each complex over time. X-axis = species type/size. Y-axis = averaged time. Z-axis = relative occurance.
        """

        return complex_time_3d(2,2,self.full_hist, self.FileNum, self.InitialTime, self.FinalTime,
                self.SpeciesName, TimeBins, xBarSize, ShowFig,
                False, False, SaveFig, SaveVars)


    def hist_3d_monomer_fraction(self,TimeBins: int, xBarSize: int = 1, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
        """Takes in a histogram.dat file from NERDSS, and creates a 3D histogram that shows % of monomers in each complex size over time.

        Args:
            TimeBins (int): The number of bins that the selected time period is divided into.
            xBarSize (int, optional): The size of each data bar in the x-dimension. Defaults to 1.
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.


        Returns:
            Returns a 3D histogram representing the aver count of each complex over time. X-axis = species type/size. Y-axis = averaged time. Z-axis = relative occurance.
        """

        return complex_time_3d(2,3,self.full_hist, self.FileNum, self.InitialTime, self.FinalTime,
                self.SpeciesName, TimeBins, xBarSize, ShowFig,
                False, False, SaveFig, SaveVars)


    ##Number of complexes over time (2d)
    def line_mean_complex_size(self, ExcludeSize: int = 0, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
        """Creates graph of the mean number of species in a single complex molecule over a time period.

        Args:
            ExcludeSize (int): Monomers in the complex that are smaller or equal to this number will not be included. 
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

        Returns:
            graph. X-axis = time. Y-axis = mean number of species in a single complex molecule.
        """
        return line_size_over_time(Data = 1, full_hist = self.full_hist, FileNum = self.FileNum, InitialTime = self.InitialTime, FinalTime = self.FinalTime,
                SpeciesName = self.SpeciesName, ExcludeSize = ExcludeSize, ShowFig = ShowFig, SaveFig = SaveFig, SaveVars = SaveVars)


    def line_max_complex_size(self, ExcludeSize: int = 0, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
        """Creates graph of the max number of species in a single complex molecule over a time period.

        Args:
            ExcludeSize (int, optional): Monomers in the complex that are smaller or equal to this number will not be included. 
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

        Returns:
            graph. X-axis = time. Y-axis = max number of species in a single complex molecule.
        """  
        return line_size_over_time(Data = 2, full_hist = self.full_hist, FileNum = self.FileNum, InitialTime = self.InitialTime, FinalTime = self.FinalTime,
                SpeciesName = self.SpeciesName, ExcludeSize = ExcludeSize, ShowFig = ShowFig, SaveFig = SaveFig, SaveVars=SaveVars)


    ##Baby Basic Histogram
    def hist_complex_count(self, BarSize: int = 1, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
        """Creates histogram of the average number of complex species that have a certain number of species.

        Args:
            ExcludeSize (int, optional): Monomers in the complex that are smaller or equal to this number will not be included. 
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.


        Returns:
            Histogram. X-axis = # of species in a complexes. Y-axis = relative count of each complex over the whole timeframe
        """
        return hist_complex_count(full_hist=self.full_hist, FileNum=self.FileNum,InitialTime= self.InitialTime, FinalTime= self.FinalTime, 
                           SpeciesName=self.SpeciesName, BarSize=BarSize, ShowFig=ShowFig, SaveFig=SaveFig, SaveVars=SaveVars)


    ##general histogram functions
    def hist_to_csv(self,OpName: int = "histogram"):
        """Creates a .csv (spreadsheet) file from a histogram.dat file (multi-species)

        Args:
            OpName (str, Optional = "histogram"): what the outputted .csv file will be named

        Returns:
            histogram.csv file: Each row is a different time stamp (all times listed in column A). Each column is a different size of complex molecule (all sizes listed in row 1). Each box 
            is the number of that complex molecule at that time stamp.
        """  
        return hist_to_csv(self.full_hist, [self.SpeciesName], OpName, True)

    def hist_to_df(self, OpName: int = "histogram", SaveCsv: bool = True):
        """Creates a pandas dataframe from a histogram.dat (multi-species)

        Args:
            OpName (str, Optional = "histogram"): what the outputted .csv file will be named
            SaveCsv (bool, optional): If a .csv file is saved as well. Defaults to True.

        Returns:
            pandas.df: Each row is a different time stamp (all times listed in column A). Each column is a different size of complex molecule (all sizes listed in row 1). Each box 
                is the number of that complex molecule at that time stamp.
        """
        return hist_to_df(self.full_hist, [self.SpeciesName], OpName, SaveCsv, True)



import math
import numpy as np

def icos_vert_norm_input(scaler: float, dis_: float):
    """Calculate normalized input coordinates for an icosahedron.

    The function calculates the normalized input coordinates for an icosahedron based on the given scaler and dis_ values.
    The scaler value is used to scale the vectors, and the dis_ value is used to determine the z-coordinate of the leg
    vectors. The resulting coordinates are returned as a tuple containing the center of mass (COM) vector, and the vectors
    for each leg, rounded to 12 decimal places.

    Args:
        scaler (float): The scaling factor for the vectors.
        dis_ (float): The z-coordinate value for the leg vectors.

    Returns:
        tuple: A tuple containing the center of mass (COM) vector, and vectors for each leg of the icosahedron.
        Each vector is represented as a numpy array of shape (3,), and is rounded to 12 decimal places.
    """
    c1 = math.cos(2*math.pi/5)
    c2 = math.cos(math.pi/5)
    s1 = math.sin(2*math.pi/5)
    s2 = math.sin(4*math.pi/5)
    v0 = scaler*np.array([0, 1])
    v1 = scaler*np.array([-s1, c1])
    v2 = scaler*np.array([-s2, -c2])
    v3 = scaler*np.array([s2, -c2])
    v4 = scaler*np.array([s1, c1])
    lg1 = np.array([v0[0], v0[1], -dis_])
    lg2 = np.array([v1[0], v1[1], -dis_])
    lg3 = np.array([v2[0], v2[1], -dis_])
    lg4 = np.array([v3[0], v3[1], -dis_])
    lg5 = np.array([v4[0], v4[1], -dis_])
    COM = np.array([0, 0, 0])
    n = np.array([0, 0, 1])
    return COM, lg1, lg2, lg3, lg4, lg5, n





from .cube_vert_norm_input import cube_vert_norm_input


def cube_vert_write(radius: float, sigma: float):
    """
    Writes input parameters for a cube vertex-centered simulation to a file.

    Args:
        radius (float): The radius of the cubes.
        sigma (float): The sigma value.


    Returns:
        parm.inp/cube.mol file: inputs for NERDSS
    """
    
    COM, lg1, lg2, lg3, n = cube_vert_norm_input(radius, sigma)
    f = open('parm.inp', 'w')
    f.write(' # Input file (cube vertex-centered)\n\n')
    f.write('start parameters\n')
    f.write('    nItr = 10000000 #iterations\n')
    f.write('    timeStep = 0.1\n')
    f.write('    timeWrite = 10000\n')
    f.write('    pdbWrite = 10000\n')
    f.write('    trajWrite = 10000\n')
    f.write('    restartWrite = 50000\n')
    f.write('    checkPoint = 1000000\n')
    f.write('    overlapSepLimit = 7.0\n')
    f.write('end parameters\n\n')
    f.write('start boundaries\n')
    f.write('    WaterBox = [500,500,500]\n')
    f.write('end boundaries\n\n')
    f.write('start molecules\n')
    f.write('    cube : 200\n')
    f.write('end molecules\n\n')
    f.write('start reactions\n')
    f.write('    cube(lg1) + cube(lg1) <-> cube(lg1!1).cube(lg1!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg2) + cube(lg2) <-> cube(lg2!1).cube(lg2!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg3) + cube(lg3) <-> cube(lg3!1).cube(lg3!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg1) + cube(lg2) <-> cube(lg1!1).cube(lg2!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg1) + cube(lg3) <-> cube(lg1!1).cube(lg3!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    cube(lg2) + cube(lg3) <-> cube(lg2!1).cube(lg3!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [M_PI, M_PI, nan, nan, 0]\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('end reactions\n')

    f = open('cube.mol', 'w')
    f.write('##\n')
    f.write('# Cube (vertex-centered) information file.\n')
    f.write('##\n\n')
    f.write('Name = cube\n')
    f.write('checkOverlap = true\n\n')
    f.write('# translational diffusion constants\n')
    f.write('D = [13.0, 13.0, 13.0]\n\n')
    f.write('# rotational diffusion constants\n')
    f.write('Dr = [0.03, 0.03, 0.03]\n\n')
    f.write('# Coordinates\n')
    f.write('COM   ' + str(round(COM[0], 8)) + '   ' +
            str(round(COM[1], 8)) + '   ' + str(round(COM[2], 8)) + '\n')
    f.write('lg1   ' + str(round(lg1[0], 8)) + '   ' +
            str(round(lg1[1], 8)) + '   ' + str(round(lg1[2], 8)) + '\n')
    f.write('lg2   ' + str(round(lg2[0], 8)) + '   ' +
            str(round(lg2[1], 8)) + '   ' + str(round(lg2[2], 8)) + '\n')
    f.write('lg3   ' + str(round(lg3[0], 8)) + '   ' +
            str(round(lg3[1], 8)) + '   ' + str(round(lg3[2], 8)) + '\n')
    f.write('\n')
    f.write('# bonds\n')
    f.write('bonds = 3\n')
    f.write('com lg1\n')
    f.write('com lg2\n')
    f.write('com lg3\n')
    f.write('\n')


# TETRAHETRON FACE AS COM




import numpy as np
import matplotlib.pyplot as plt
from ...file_managment.save_vars_to_file import save_vars_to_file


def hist_complex_count(full_hist: list, FileNum: int, InitialTime: float, FinalTime: float, SpeciesName: str,
         BarSize: int = 1, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
    """Creates histogram of the average number of complex species that have a certain number of species.

    Args:
        full_hist (list): List that holds all of the data from the histogram.dat file
        FileNum (int): Number of the total input files (file names should be [fileName]_1,[fileName]_2,...)
        InitialTime (float): The starting time. Must not be smaller / larger then times in file.
        FinalTime (float): The ending time. Must not be smaller / larger then times in file.
        SpeciesName (str): The name of the species you want to examine. Should be in the .dat file.
        BarSize (int, optional): The size of each data bar in the x-dimension. Defaults to 1.
        ShowFig (bool, optional): If the plot is shown. Defaults to True.
        SaveFig (bool, optional): If the plot is saved. Defaults to False.
        SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

    Returns:
        Histogram. X-axis = # of species in a complexes. Y-axis = relative count of each complex over the whole timeframe
    """

    
    #main lists, each sublist = 1 folder
    count_list = []
    size_list = []
    
    #runs through each file
    for hist in full_hist:

        #lists for this file
        total_size_list = [] #list of each size
        total_count_list = [] #list of lists that hold the counts of eachsize

        data_count = 0
        for timestep in hist:
            if InitialTime <= timestep[0] <= FinalTime:
                data_count += 1
                
                #runs through each size of complex in this timestep
                for complex_index,complex_size in enumerate(timestep[2]):
                    
                    #if it is not in the total size list, add it
                    if complex_size not in total_size_list:
                        total_size_list.append(complex_size)
                        total_count_list.append(timestep[1][complex_index])
                    
                    #if it is, only add the number of this size to the list
                    else:
                        index = total_size_list.index(complex_size)
                        total_count_list[index] += timestep[1][complex_index]
        
        #entire array devided by datacount to determine % of the complexes that are that size
        total_count_list = np.array(total_count_list)/data_count
        
        #sort the size list
        if len(total_size_list) != 0:
            total_size_list_sorted = np.arange(1, max(total_size_list)+1, 1)
        else:
            total_size_list_sorted = np.array([])
        
        #sort the count list
        total_count_list_sorted = []
        for size in total_size_list_sorted:
            if size in total_size_list:
                index = total_size_list.index(size)
                total_count_list_sorted.append(total_count_list[index])
            else:
                total_count_list_sorted.append(0.0)
        
        #append sizes/counts to the main list
        size_list.append(total_size_list_sorted)
        count_list.append(total_count_list_sorted)
    
    #determine the biggest possible complex sizes, and make an empty array of that size
    max_size = 0
    for size_file in size_list:
        if max_size < len(size_file):
            max_size = len(size_file)
            n_list = size_file
    count_list_filled = np.zeros([max_size, FileNum])
    
    #add counts to the 0.0 list and make it so each sublist = 1 size across all files, instead of each sublist being 1 file with all sizes
    for i in range(len(count_list)):
        for j in range(len(count_list[i])):
            count_list_filled[j][i] += count_list[i][j]

    mean_ = []
    std_ = []
    n_list_ = []
    mean_temp = 0
    std_temp = 0
    bar_size_count = 0
    #run through each count list, find mean/stddev and add it to main list based on bar size
    for index,count_list in enumerate(count_list_filled):
        
        #determine mean / std for this count_list
        if FileNum != 1:
            mean_temp += np.nanmean(count_list)
            std_temp += np.nanstd(count_list)  
        else:
            mean_temp += count_list[0]

        #determine whether or not to add it based on bar size
        bar_size_count += 1
        
        #if it is the last run
        if index+1 == len(count_list_filled):
            mean_.append(mean_temp)
            std_.append(std_temp)
            n_list_.append(n_list[index])
        
        #if enough sizes have been gone through to create a new bar
        elif bar_size_count >= BarSize:
            mean_.append(mean_temp)
            std_.append(std_temp)
            n_list_.append(n_list[index])
            mean_temp = 0
            std_temp = 0
            bar_size_count = 0    
    
    mean_ = np.array(mean_)
    std_ = np.array(std_)
    n_list_ = np.array(n_list_)
    
    #save vars
    if SaveVars:
        save_vars_to_file({"cmplx_sizes":n_list_, "mean":mean_, "std":std_})


    #show figure!
    if ShowFig:
        if FileNum != 1:
            plt.bar(n_list_, mean_, width=BarSize, color='C0',
                    yerr=std_, ecolor='C1', capsize=2)
        else:
            plt.bar(n_list_, mean_, width=BarSize)
        plt.title('Histogram of ' + str(SpeciesName))
        plt.xlabel('Number of ' + SpeciesName + ' in sigle complex')
        plt.ylabel('Count')
        if SaveFig:
            plt.savefig('Histogram.png', dpi=500)
        plt.show()
    return n_list_, mean_, std_





from .icos_vert_write import icos_vert_write


def icos_vert(radius: float, sigma: float):
    """Generate vertices for an icosahedron and write to a file.

    This function generates the vertices of an icosahedron with the given
    radius and sigma, and writes them to a file using the icos_vert_write
    function from the .icos_vert_write module. After writing is complete,
    it prints a message indicating the file writing status.

    Args:
        radius (float): The radius of the icosahedron.
        sigma (float): The sigma value used in the generation of vertices.

    Returns:
        parm.inp/icos.mol: input files for NERDSS

    Example:
        >>> icos_vert(2.0, 0.5)
        File writing complete!
    """
    icos_vert_write(radius, sigma)
    print('File writing complete!')
    return 0


# -----------------------------------Data Visualization------------------------------

# Analysis tools for 'histogram_complexes_time.dat' file





from .tetr_vert_write import tetr_vert_write


def tetr_vert(radius: float, sigma: float):
    """Writes tetrahedron vertices to a file.
    
    Args:
        radius (float): The radius of the tetrahedron's circumsphere.
        sigma (float): The height of the tetrahedron.
        
    Returns:
        parm.inp/icos.mol: input files for NERDSS
                
    Example:
        >>> tetr_vert(1.0, 0.5)
        File writing complete!
    """
    tetr_vert_write(radius, sigma)
    print('File writing complete!')
    return 0





import math


def real_PDB_chain_int(unique_chain, split_position, split_resi_count, split_atom_count, split_resi_type, split_atom_type, split_resi_position):
    """
    This function takes a complex protein structure and determines which chains and residues are interacting
    with each other based on the distance between atoms. The output is a tuple that includes the following lists:

    Args:
        unique_chain (list): Unique chains within the protein structure
        split_position (list of list): Each sublist contains the positions of atoms in a specific chain
        split_resi_count (list of list): Each sublist contains the residue count of atoms in a specific chain
        split_atom_count (list of list): Each sublist contains the atom count in a specific chain
        split_resi_type (list of list): aEach sublist contains the residue type of atoms in a specific chain
        split_atom_type (list of list): Each sublist contains the atom type of atoms in a specific chain
        split_resi_position (list of list): Each sublist contains the residue position of atoms in a specific chain
    
    Returns:
        reaction_chain (list of list): Each sublist contains the chain IDs of two chains that are interacting with each other
        reaction_resi_count (list of list of list): Each sub-sublist contains pairs of residue counts that are interacting
        reaction_resi_type (list of list of list): Each sub-sublist contains pairs of residue types that are interacting
        reaction_resi_position (list of list of list): Each sub-sublist contains pairs of residue positions that are interacting
        reaction_atom (list of list): Each sublist contains pairs of atom counts that are interacting
        reaction_atom_position (list of list): Each sublist contains pairs of atom positions that are interacting
        reaction_atom_distance (list): a list of distances between interacting atoms
        reaction_atom_type (list of list): Each sublist contains pairs of atom types that are interacting
    """

    distance = 0
    # list of lists (each sublist will include two letters indicating these two chains have
    reaction_chain = []
    # interaction) eg: in this protein, only chain A&B, A&D and C&D are interacting, then the list will look like
    # [[A,B],[A,D],[C,D]]
    # list of lists of lists(each sub-sublist will include a bunch of lists of residue pairs
    reaction_resi_type = []
    # (without repeats)) eg: [[[resia,resib],[resic,resid]],[[resie,resif],[resig,resih]],[[resii,resij],[resik,resil]]]
    # ----reaction residues of chain-------- A&B------------------------A&D-------------------------C&D -------------
    reaction_resi_count = []
    reaction_atom = []
    reaction_atom_position = []
    reaction_atom_distance = []
    reaction_atom_type = []
    reaction_resi_position = []
    for i in range(len(unique_chain) - 1):
        for j in range(i+1, len(unique_chain)):
            inner_atom_position = []
            inner_atom_distance = []
            inner_atom = []
            inner_reaction_resi_count = []
            inner_reaction_resi_type = []
            inner_reaction_atom_type = []
            inner_reaction_resi_position = []
            for m in range(len(split_position[i])):
                for n in range(len(split_position[j])):
                    distance = math.sqrt((split_position[i][m][0]-split_position[j][n][0])**2
                                         + (split_position[i][m][1]-split_position[j][n][1])**2
                                         + (split_position[i][m][2]-split_position[j][n][2])**2)
                    if distance <= 0.3:
                        inner_atom.append(
                            [split_atom_count[i][m], split_atom_count[j][n]])
                        inner_atom_distance.append(distance)
                        inner_atom_position.append(
                            [split_position[i][m], split_position[j][n]])
                        inner_reaction_atom_type.append(
                            [split_atom_type[i][m], split_atom_type[j][n]])
                        if [split_resi_count[i][m], split_resi_count[j][n]] not in inner_reaction_resi_count:
                            inner_reaction_resi_count.append(
                                [split_resi_count[i][m], split_resi_count[j][n]])
                            inner_reaction_resi_position.append(
                                [split_resi_position[i][m], split_resi_position[j][n]])
                            inner_reaction_resi_type.append(
                                [split_resi_type[i][m], split_resi_type[j][n]])
            if len(inner_reaction_resi_count) > 0:
                reaction_chain.append([unique_chain[i], unique_chain[j]])
                reaction_resi_count.append(inner_reaction_resi_count)
                reaction_resi_type.append(inner_reaction_resi_type)
                reaction_atom.append(inner_atom)
                reaction_atom_position.append(inner_atom_position)
                reaction_atom_distance.append(inner_atom_distance)
                reaction_atom_type.append(inner_reaction_atom_type)
                reaction_resi_position.append(inner_reaction_resi_position)
    return reaction_chain, reaction_atom, reaction_atom_position, reaction_atom_distance, reaction_resi_count, \
        reaction_resi_type, reaction_atom_type, reaction_resi_position





import math
from ..gen_platonic.distance import distance


def dode_face_leg_reduce(COM: float, leg: float, sigma: float):
    """Calculates the reduced length of a dodecahedron leg based on the given center of mass (COM), leg coordinates,
    and sigma value.

    Args:
        COM (float): The coordinates of the center of mass as a list [x, y, z], where x, y, and z are floats.
        leg (float): The coordinates of the leg as a list [x, y, z], where x, y, and z are floats.
        sigma (float): The sigma value for the dodecahedron face.

    Returns:
        list: A list containing the reduced coordinates of the leg after applying the reduction factor.
        The list contains three elements [x', y', z'], where x', y', and z' are the reduced coordinates of the leg
        rounded to 14 decimal places.
    """
    
    # calculate the recuced length when considering the sigma value
    n = 14
    m = (1+5**(0.5))/2
    angle = 2*math.atan(m)
    red_len = sigma/(2*math.sin(angle/2))
    ratio = 1 - red_len/distance(COM, leg)
    leg_red = []
    for i in range(0, 3):
        leg_red.append(round((leg[i] - COM[i])*ratio + COM[i], n))
    return leg_red





import numpy as np
import matplotlib.pyplot as plt
from .read_multi_hist import read_multi_hist
from ...file_managment.save_vars_to_file import save_vars_to_file

def multi_stack_hist_complex_count(FullHist: list, FileNum: int, InitialTime: float, FinalTime: float,
                       SpeciesList: list, xAxis: str, DivideSpecies: str, DivideSize: int,
                       BarSize: int = 1, ExcludeSize: int = 0, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
    """Creates a stacked histogram from histogram.dat (multi-species) that shows the average number of each type of 
    complex species (based on protein composition) over the whole sim. 

    Args:
        FullHist (list): Holds all of the information from the .dat file
        FileNum (int): Number of the total input files (file names should be [fileName]_1,[fileName]_2,...)
        InitialTime (float): The starting time. Must not be smaller / larger then times in file.
        FinalTime (float): The ending time. Must not be smaller / larger then times in file.
        xAxis (str): Species shown on X-axis.
        DivideSpecies (str): The name of the species that will be seperated by size.
        DivideSize (int): The value that separates the size of dissociate complexes. (only changes color of graph)
        SpeciesList (list): The names of the species you want to examine. Should be in the .dat file.
        BarSize (int, optional): The size of each data bar in the X-dimension. Defaults to 1.
        ExcludeSize (int, optional): Monomers in the complex that are smaller or equal to this number will not be included. 
        ShowFig (bool, optional): If the plot is shown. Defaults to True.
        SaveFig (bool, optional): If the plot is saved. Defaults to False.
        SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

    Returns:
        Histogram. X-axis = size of selected species, Y-axis = average number of each corresponds.
    """

    above_list = []
    equal_list = []
    below_list = []
    max_size = 0 # largest complex size

    #get index of the xAxis and divide species
    x_species_index = SpeciesList.index(xAxis)
    divide_species_index = SpeciesList.index(DivideSpecies)

    #for each file
    for hist_list in FullHist:

        total_above_dict = {}
        total_equal_dict = {}
        total_below_dict = {}

        data_count = 0

        #for every time step
        for time_step in hist_list:
            if time_step != []:
                time = time_step[0]
                if InitialTime <= time <= FinalTime:
                    data_count += 1

                    #for every protein complex
                    for protein_complex in time_step[1:]:
                        
                        #if number of selected proteisn in the complex
                        if xAxis == 'tot' and DivideSpecies in SpeciesList:
                            total_size = sum(protein_complex[0:-1])
                        elif xAxis in SpeciesList and DivideSpecies in SpeciesList:
                            total_size = protein_complex[x_species_index]
                        
                        if total_size >= ExcludeSize:
                            divide_spe_size = protein_complex[divide_species_index]
                                
                            #check if there is already a key for a protein of this size. If there isn't add it
                            if not total_size in total_above_dict:
                                total_above_dict[total_size] = 0
                                total_equal_dict[total_size] = 0
                                total_below_dict[total_size] = 0

                            #if complex size not included, initlize this size, and add the total number of this kind of complex to lists of above/below
                            if divide_spe_size > DivideSize:
                                total_above_dict[total_size] += (protein_complex[-1])
                            elif divide_spe_size == DivideSize:
                                total_equal_dict[total_size] += (protein_complex[-1])
                            else:
                                total_below_dict[total_size] += (protein_complex[-1])
                            
        #find max size and devide each protein by # of timesteps to get average
        for key in total_above_dict:
            total_above_dict[key] = total_above_dict[key] / data_count
            if max_size < int(key):
                max_size = int(key)
            n_list = list(range(1,int(max_size+1)))
        
        for key in total_equal_dict:
            total_equal_dict[key] = total_equal_dict[key] / data_count
            if max_size < int(key):
                max_size = int(key)
            n_list = list(range(1,int(max_size+1)))
        
        for key in total_below_dict:
            total_below_dict[key] = total_below_dict[key] / data_count
            if max_size < int(key):
                max_size = int(key)
            n_list = list(range(1,int(max_size+1)))

        #add dictionaries to main lists
        above_list.append(total_above_dict)
        equal_list.append(total_equal_dict)
        below_list.append(total_below_dict)


    #add dictionary values to filled lists and transposes them for prep for mean
    above_list_filled = np.zeros([max_size,FileNum])
    for indexX,dict in enumerate(above_list):
        for key in dict:
            above_list_filled[int(key)-1][indexX] += dict[key]

    equal_list_filled = np.zeros([max_size,FileNum])
    for indexX,dict in enumerate(equal_list):
        for key in dict:
            equal_list_filled[int(key)-1][indexX] += dict[key]
    
    below_list_filled = np.zeros([max_size,FileNum])
    for indexX,dict in enumerate(below_list):
        for key in dict:
            below_list_filled[int(key)-1][indexX] += dict[key]
    
    #is it mean time?
    mean_above = []
    std_above = []
    mean_equal = []
    std_equal = []
    mean_below = []
    std_below = []
    for protein_size in above_list_filled:
        mean_above.append(np.nanmean(protein_size))
        std_above.append(np.nanstd(protein_size))
    for protein_size in equal_list_filled:
        mean_equal.append(np.nanmean(protein_size))
        std_equal.append(np.nanstd(protein_size))
    for protein_size in below_list_filled:
        mean_below.append(np.nanmean(protein_size))
        std_below.append(np.nanstd(protein_size))
    

    
    #will combine means together if bar size is too small
    mean_above_ = []
    mean_equal_ = []
    mean_below_ = []
    std_above_ = []
    std_equal_ = []
    std_below_ = []
    n_list_ = []
    temp_mean_above = 0
    temp_mean_equal = 0
    temp_mean_below = 0
    temp_std_above = 0
    temp_std_equal = 0
    temp_std_below = 0
    bar_size_count = 0
    for i in range(len(mean_above)):
        temp_mean_above += mean_above[i]
        temp_mean_equal += mean_equal[i]
        temp_mean_below += mean_below[i]
        temp_std_above += std_above[i]
        temp_std_equal += std_equal[i]
        temp_std_below += std_below[i]
        bar_size_count += 1
        if bar_size_count >= BarSize and i != len(mean_above) - 1:
            mean_above_.append(temp_mean_above)
            mean_equal_.append(temp_mean_equal)
            mean_below_.append(temp_mean_below)
            std_above_.append(temp_std_above)
            std_equal_.append(temp_std_equal)
            std_below_.append(temp_std_below)
            n_list_.append(n_list[i])
            temp_mean_above = 0
            temp_mean_equal = 0
            temp_mean_below = 0
            temp_std_above = 0
            temp_std_equal = 0
            temp_std_below = 0
            bar_size_count = 0
    
    mean_above_.append(temp_mean_above)
    mean_equal_.append(temp_mean_equal)
    mean_below_.append(temp_mean_below)
    std_above_.append(temp_std_above)
    std_equal_.append(temp_std_equal)
    std_below_.append(temp_std_below)
    n_list_.append(n_list[i])
    mean_above_ = np.array(mean_above_)
    mean_equal_ = np.array(mean_equal_)
    mean_below_ = np.array(mean_below_)
    std_above_ = np.array(std_above_)
    std_equal_ = np.array(std_equal_)
    std_below_ = np.array(std_below_)
    n_list_ = np.array(n_list_)
    
    #output variables
    if SaveVars:
        save_vars_to_file({"x_mono_count":n_list_, "cmplx_count":[mean_below_, mean_equal_, mean_above_], "std":[std_below_, std_equal_, std_above_]})
    #show figure!
    if ShowFig:
        if DivideSize != 0:
            below_label = DivideSpecies + '<' + str(DivideSize)
            equal_label = DivideSpecies + '=' + str(DivideSize)
            above_label = DivideSpecies + '>' + str(DivideSize)
        else:
            above_label = 'With ' + DivideSpecies
            equal_label = 'Without ' + DivideSpecies
        if FileNum != 1:
            if DivideSize != 0:
                plt.bar(n_list_, mean_below_, width=BarSize, color='C0',
                        yerr=std_below_, label=below_label, ecolor='C3', capsize=2)
            plt.bar(n_list_, mean_equal_, width=BarSize, color='C1', yerr=std_equal_,
                    bottom=mean_below_, label=equal_label, ecolor='C3', capsize=2)
            plt.bar(n_list_, mean_above_, width=BarSize, color='C2', yerr=std_above_,
                    bottom=mean_below_+mean_equal_, label=above_label, ecolor='C3', capsize=2)
        else:
            if DivideSize != 0:
                plt.bar(n_list_, mean_below_, width=BarSize,
                        color='C0', label=below_label, capsize=2)
            plt.bar(n_list_, mean_equal_, width=BarSize, color='C1',
                    bottom=mean_below_, label=equal_label, capsize=2)
            plt.bar(n_list_, mean_above_, width=BarSize, color='C2',
                    bottom=mean_below_+mean_equal_, label=above_label, capsize=2)
        if xAxis == 'tot':
            x_label_name = 'total monomers'
        else:
            x_label_name = xAxis
        plt.xlabel('Number of ' + x_label_name + ' in sigle complex')
        plt.ylabel('Count')
        plt.legend()
        plt.title('Histogram of Multi-component Assemblies')
        fig_name = 'stacked_histogram_of_' + xAxis + '_divided_by_' + DivideSpecies
        if SaveFig:
            plt.savefig(fig_name, dpi=500)
        plt.show()
    return n_list_, [mean_below_, mean_equal_, mean_above_], [std_below_, std_equal_, std_above_]


